/***************************** covar.c ***********************************
*	UTILITY SUBROUTINES FOR COVAR STRUCTURES
*	Jeff Fessler
*	COPYRIGHT 1989,  THE BOARD OF TRUSTEES OF THE
*	LELAND STANFORD JUNIOR UNIVERSITY.  ALL RIGHTS RESERVED.
*************************************************************************/
#include "smooth.h"


/****************************** inv_to_covar() ***************************
*	CONVERT INVERSE OF COVARAINCE MATRIX INTO COVAR STRUCTURE
*	INPUT:
*		inv: CMAT[1..MxM]
*	OUTPUT:
*		cov:
*	RETURNS:	1 if OK, 0 if singular
*************************************************************************/
static bool inv_to_covar(COVAR *cov, CMAT inv, int M)
{
	mat_to_sym_band(cov->inv, inv, M-1, M);
	if ( !chol_sym_band(cov->Linv, cov->Dinv, cov->inv, M-1, M) ) return Failure;
	invert_sym_band(cov->cov, cov->Linv, cov->Dinv, M-1, M);
	cov->trace = band_trace(cov->cov, M-1, M);
	return Success;
}


/*
*	ALLOCATE SPACE FOR A SET OF N COVARIANCE STRUCTURES
*	WITH COVARIANCE SIZE MxM
*/
COVAR *alloc_covar(int M, int N)
{
	COVAR *cov;
	Alloc0(cov, COVAR, N, sizeof(COVAR))
	cov--;
	
    for (; N >= 1; N--) {
	if ( !(cov[N].cov = ALLOC_MATRIX(M, M)) )	return (COVAR *) 0;
	if ( !(cov[N].cov2 = ALLOC_MATRIX(M, M)) )	return (COVAR *) 0;
	if ( !(cov[N].inv = ALLOC_MATRIX(M, M)) )	return (COVAR *) 0;
	if ( !(cov[N].Linv = ALLOC_MATRIX(M, M)) )	return (COVAR *) 0;
	if ( !(cov[N].Dinv = ALLOC_VECTOR(M)) )		return (COVAR *) 0;
    }
	return cov;
}


void covar_free(COVAR *cov, int N)
{
    for (; N >= 1; N--) {
	FREE_MATRIX(cov[N].cov);
	FREE_MATRIX(cov[N].cov2);
	FREE_MATRIX(cov[N].inv);
	FREE_MATRIX(cov[N].Linv);
	FREE_VECTOR(cov[N].Dinv);
    }
    Free(cov+1);
}


void copy_covar(COVAR *src, COVAR *dst, int MM)
{
	COPY_MATRIX(src->cov, dst->cov, MM, MM);
	COPY_MATRIX(src->cov2, dst->cov2, MM, MM);
	COPY_MATRIX(src->inv, dst->inv, MM, MM);
	COPY_MATRIX(src->Linv, dst->Linv, MM, MM);
	COPY_VECTOR(src->Dinv, dst->Dinv, MM);
}


/*
*	MULTIPLY covariance by scale - correcting others
*/
void scale_covar(COVAR *cov, double scale, int MM)
{
	SCALE_MATRIX(cov->cov, scale, MM, MM);
	SCALE_MATRIX(cov->cov2, scale*scale, MM, MM);
	SCALE_MATRIX(cov->inv, 1. / scale, MM, MM);
	SCALE_VECTOR(cov->Dinv, 1. / scale, MM);
}


/****************************** inv_to_covar_set() ***********************
*	CONVERT SET OF COVARIANCE MATRIX INVERSES INTO COVAR STRUCTURE
*	INPUT:
*		isigma[1..MxMx(Nc)]		covariance inverses
*		flag_iid			Nc = 1 or N ?
*	RETURN:
*		cov[1..(Nc)]:			or 0 if problem
*************************************************************************/
COVAR *inv_to_covar_set(CMAT isigma, int M, int N, bool flag_iid)
{
	int nn, Nc = flag_iid ? 1 : N;
	COVAR *covar;

	if ( !(covar = alloc_covar(M, Nc)) )	return 0;

	for (nn=Nc; nn >= 1; nn--)
		if (!inv_to_covar(&covar[nn], isigma+(nn-1)*M*M, M)) return 0;

	return covar;
}
