/*********************** example1.c **************************************
*	EXAMPLE FRONT END ROUTINE FOR USING VECTOR-SPLINE SUBROUTINES TO
*	SMOOTH VECTOR MEASUREMENTS, COMPUTE CV SCORES, ETC.
*	Jeff Fessler
*	COPYRIGHT 1989,  THE BOARD OF TRUSTEES OF THE
*	LELAND STANFORD JUNIOR UNIVERSITY.  ALL RIGHTS RESERVED.
*************************************************************************/
#include "smooth.h"


/*
*	VECTOR SPLINE SMOOTH AND COMPUTE SCORES
*	INPUT:
*		alpha[1..M]		smoothing parameters
*		ti[1..N]		sample points
*		yi[1..nxM]		measurements
*		isigma[1..MxMx(N)]	inverse of measurement error covariances
*		flag_iid		0: one covariance, 1: N covariances
*	OUTPUT:
*		ys[1..nxM]		smoothed estimates
*	RETURNS:			1 if successful, 0 if error
*/
bool vspline(CMAT ys, ALPHA_SET alpha, VECT ti, CMAT yi, CMAT isigma, bool flag_iid, int MM, int nn)
{
	COVAR *covar;

	/* convert the flattened covariance inverses to structure	*/
	if ( !(covar = inv_to_covar_set(isigma, MM, nn, flag_iid)) )	return Failure;
	if ( !vs_setup_alloc(MM, nn))				return Failure;
	if ( !vs_setup_ti(ti, nn))				return Failure;
	if ( !vs_setup_yi(yi, MM, nn))				return Failure;
	if ( !vs_setup_cov(covar, flag_iid))			return Failure;
	if ( !vs_setup_alpha(alpha))				return Failure;
	if ( !vs_smooth(ys, 1) )				return Failure;
	vs_cleanup();
	covar_free(covar, flag_iid ? 1 : nn);

	return Success;
}


/*
*	VECTOR SPLINE SMOOTH AND COMPUTE SCORES
*	INPUT:
*		alpha[1..M]		smoothing parameters
*		ti[1..N]		sample points
*		yi[1..nxM]		measurements
*		covar[1..?]		measurement error covariances
*		flag_iid		covariance varies?
*	OUTPUT:
*		ys[1..nxM]		smoothed estimates
*		ur			unbiased risk score
*		cv			cross validation score
*		gcv			generalized CV score
*		penin			initial roughness penalty
*		penout			final roughness penalty
*		rough			final roughness penalty
*/
bool vect_smooth(CMAT ys, double *ur, double *cv, double *gcv,
double *penin, double *penout, double *rough, ALPHA_SET alpha,
VECT ti, CMAT yi, COVAR *covar, bool flag_iid, int MM, int nn)
{
	/* SETUP: INDEPENDENT OF ALPHA 					*/
	if ( !vs_setup_alloc(MM, nn))				return Failure;
	if ( !vs_setup_ti(ti, nn))				return Failure;
	if ( !vs_setup_yi(yi, MM, nn))				return Failure;
	if ( !vs_setup_cov(covar, flag_iid))			return Failure;
	if ( !cv_setup_alloc())					return Failure;
	if ( !cv_setup_covar())					return Failure;

	/* THE ROUTINES BELOW ALL DEPEND ON ALPHA			*/
	if ( !vs_setup_alpha(alpha))				return Failure;
	if ( !vs_smooth(ys, 1))					return Failure;
	if ( !cross_val(ur, cv, gcv, JOBUR | JOBCV | JOBGCV))	return Failure;

	*rough = roughness();
	if ( !penalty(penout, alpha, ys, MM, nn) )		return Failure;
	if ( !penalty(penin, alpha, yi, MM, nn) )		return Failure;

	vs_cleanup();
	cv_cleanup();
	return Success;
}
